using System;
using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Threading;
using MVR;
using MeshVR;
using System.Linq;
using MVR.FileManagementSecure;


namespace Stopper {
	public class Contortionist : MVRScript {
		
		string savePath = "Saves\\Contortionist";
		string fileExt = "jointconfig";
		
		List<string> groups = new List<string>{"Arms and Shoulders", "Legs", "Torso", "Head", "Fingers", "Toes"};
		
		Dictionary<string,string> jointNameToGroup = new Dictionary<string,string>{
			["head"] = "Head",
			["neck"] = "Head",
			["lowerJaw"] = "Head",
			
			
			["collar"] = "Arms and Shoulders",
			["shldr"] = "Arms and Shoulders",
			["forearm"] = "Arms and Shoulders",
			["hand"] = "Arms and Shoulders",
			
			["chest"] = "Torso",
			["abdomen"] = "Torso",
			["hip"] = "Torso",
			["pelvis"] = "Torso",
			
			["thigh"] = "Legs",
			["shin"] = "Legs",
			["foot"] = "Legs",
			
			
			["pinky"] = "Fingers",
			["ring"] = "Fingers",
			["mid"] = "Fingers",
			["index"] = "Fingers",
			["thumb"] = "Fingers",
			
			["toe"] = "Toes",
		};
		
		
		Dictionary<JointConfig, JointConfig> jointSymmetryMapping = new Dictionary<JointConfig, JointConfig>();
		Dictionary<string, List<JointConfig>> groupToJoints = new Dictionary<string, List<JointConfig>>();
		
		List<JSONStorableFloat> limits = new List<JSONStorableFloat>();
		
		List<JointConfig> jointConfigs = new List<JointConfig>();
		
		public JSONStorableBool symmetricMode;
		
		public class JointConfig {
			public Contortionist script;
			public string name;
			public string symmetryName;
			public string group;
			public ConfigurableJoint joint;
			public JointConfig symmetricJoint;
			
			public JSONStorableFloat lowAngularXLimit;
			public JSONStorableFloat highAngularXLimit;
			public JSONStorableFloat angularYLimit;
			public JSONStorableFloat angularZLimit;
			
			public UIDynamicSlider lowAngularXLimitSlider;
			public UIDynamicSlider highAngularXLimitSlider;
			public UIDynamicSlider angularYLimitSlider;
			public UIDynamicSlider angularZLimitSlider;
			
			private bool uiLoaded = false;
			public bool visible = false;
			public bool ignoreSymmetry = false;
			
			
			public void Default(){
				ignoreSymmetry = true;
				lowAngularXLimit.SetValToDefault();
				highAngularXLimit.SetValToDefault();
				angularYLimit.SetValToDefault();
				angularZLimit.SetValToDefault();
				ignoreSymmetry = false;
			}
			
			public void Unconstrain(){
				ignoreSymmetry = true;
				lowAngularXLimit.val = -180f;
				highAngularXLimit.val = 180f;
				angularYLimit.val = 180f;
				angularZLimit.val = 180f;
				ignoreSymmetry = false;
			}
			
			public void Show(bool show){
				visible = show;
				if(!uiLoaded && show){
					lowAngularXLimitSlider = script.AddSlider(lowAngularXLimit,
					(float angle) => {
						var limit = joint.lowAngularXLimit;
						limit.limit = angle;
						joint.lowAngularXLimit = limit;
						
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.lowAngularXLimit.val != angle){
							symmetricJoint.lowAngularXLimit.val = angle;
						}
					}, true);
				
					highAngularXLimitSlider = script.AddSlider(highAngularXLimit,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.highAngularXLimit.val != angle){
							symmetricJoint.highAngularXLimit.val = angle;
						}
					}, true);
				
					angularYLimitSlider = script.AddSlider(angularYLimit,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.angularYLimit.val != angle){
							symmetricJoint.angularYLimit.val = angle;
						}
					}, true);
				
					angularZLimitSlider = script.AddSlider(angularZLimit,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.angularZLimit.val != angle){
							symmetricJoint.angularZLimit.val = angle;
						}
					}, true);
					uiLoaded = true;
					return;
				}
				
				if(uiLoaded){
					lowAngularXLimitSlider.gameObject.SetActive(show);
					highAngularXLimitSlider.gameObject.SetActive(show);
					angularYLimitSlider.gameObject.SetActive(show);
					angularZLimitSlider.gameObject.SetActive(show);
				}
			}
		}
		
		public override void Init() {
			
			var geometry = containingAtom.gameObject.transform.Find("rescale2/PhysicsModel");
			
			if(geometry == null)
				return;
			
			// Preset code was adapted from AcidBubbles' Collider Editor
			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Contortionlist Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});
			
			var unconstrainAllButton = CreateButton("Unconstrain Shown", true);
			unconstrainAllButton.button.onClick.AddListener(() => {
				foreach(var jointConfig in jointConfigs){
					if(jointConfig.visible){
						jointConfig.Unconstrain();
					}
				}
			});
			
			var defaultAllButton = CreateButton("Default Shown", true);
			defaultAllButton.button.onClick.AddListener(() => {
				foreach(var jointConfig in jointConfigs){
					if(jointConfig.visible){
						jointConfig.Default();
					}
				}
			});
			
			
			symmetricMode = new JSONStorableBool("Symmetric Mode", true);
			var symmetricModeToggle = CreateToggle(symmetricMode, false);
			
			foreach(Rigidbody rb in geometry.gameObject.GetComponentsInChildren<Rigidbody>()){
				
				if(rb.name.ToLower().Contains("collider"))
					continue;
				
				bool configurable = false;
				
				string group = "";
				
				foreach(var entry in jointNameToGroup){
					string jointName = entry.Key;
					if(rb.name.ToLower().Contains(jointName)){
						configurable = true;
						group = entry.Value;
					}
				}
				
				if(!configurable) 
					continue;
				
				var joint = rb.GetComponent<ConfigurableJoint>();
				if(joint == null)
					continue;
				
				JointConfig jointConfig = new JointConfig();
				jointConfig.script = this;
				
				jointConfig.group = group;
				
				jointConfig.name = rb.name;
				jointConfig.symmetryName = joint.name.ToLower().Replace("r", "x").Replace("l", "x");
				
				jointConfig.joint = joint;
				
				jointConfig.lowAngularXLimit = CreateFloat($"{rb.name} lowAngularXLimit", joint.lowAngularXLimit.limit, -180f, 180f);
				
				jointConfig.highAngularXLimit = CreateFloat($"{rb.name} highAngularXLimit", joint.highAngularXLimit.limit, -180f, 180f);
				
				jointConfig.angularYLimit = CreateFloat($"{rb.name} angularYLimit", joint.angularYLimit.limit, -180f, 180f);
				
				jointConfig.angularZLimit = CreateFloat($"{rb.name} angularZLimit", joint.angularZLimit.limit, -180f, 180f);

				jointConfig.lowAngularXLimit.setCallbackFunction = (float angle) => {
					var limit = jointConfig.joint.lowAngularXLimit;
					limit.limit = angle;
					jointConfig.joint.lowAngularXLimit = limit;
				};

				jointConfig.highAngularXLimit.setCallbackFunction = (float angle) => {
					var limit = jointConfig.joint.highAngularXLimit;
					limit.limit = angle;
					jointConfig.joint.highAngularXLimit = limit;
				};

				jointConfig.lowAngularXLimit.setCallbackFunction = (float angle) => {
					var limit = jointConfig.joint.lowAngularXLimit;
					limit.limit = angle;
					jointConfig.joint.angularZLimit = limit;
				};

				jointConfig.angularZLimit.setCallbackFunction = (float angle) => {
					var limit = jointConfig.joint.angularZLimit;
					limit.limit = angle;
					jointConfig.joint.angularZLimit = limit;
				};
				
				if(!groupToJoints.ContainsKey(group)){
					groupToJoints[group] = new List<JointConfig>();
				}
				groupToJoints[group].Add(jointConfig);
				
				
				var symmetricJoint = jointConfigs.Find((j) => j.symmetryName == jointConfig.symmetryName);
				if(symmetricJoint != null){
					jointConfig.symmetricJoint = symmetricJoint; 
					symmetricJoint.symmetricJoint = jointConfig; 
				}
				jointConfigs.Add(jointConfig);
			}
			
			jointConfigs.Sort((a,b) => a.name.CompareTo(b.name));
			
			foreach(var group in groups){
				CreateGroupButton(group);
			}
		}

		private void AdjustJoint(JointConfig jointConfig){
			
		}
		
		private void UnconstrainAll(){
			foreach(var jointConfig in jointConfigs){
				jointConfig.Unconstrain();
			}
		}
		
		private void DefaultAll(){
			foreach(var jointConfig in jointConfigs){
				jointConfig.Default();
			}
		}
		
		public void CreateGroupButton(string group){
			UIDynamicButton groupButton = CreateButton(group);
			groupButton.button.onClick.AddListener(() => {
				foreach(var jointConfig in jointConfigs){
					bool show = jointConfig.group == groupButton.label;
										
					jointConfig.Show(show);
				}
			});
		}
		
		public JSONStorableFloat CreateFloat(string name, float defaultVal, float min, float max) {
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			limits.Add(storableFloat);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			RegisterFloat(storableFloat);
			return storableFloat;
		}
		
		public UIDynamicSlider AddSlider(JSONStorableFloat storableFloat, UnityAction<float> onChange = null, bool rightSide = false)
		{
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = true;
			
			if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return uiSlider;
		}
		
		private void Start(){
			
		}
	
			
		private void OnDisable()
		{
			DefaultAll();
		}
		
		private void OnDestroy(){
		}
		
		private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);
			
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}
		
		private string GetPathToObject(GameObject go){
			 if (go.transform.parent == null)
				 return "/" + go.name;
			 return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
		}
	}
}
